/*
        xz
                Copyright 2012 by Christian Ghisler (support@ghisler.com)
                This plugin is Freeware

        to be used with Total Commander (www.ghisler.com)
*/

#include <windows.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <io.h>
#include "zlib.h"

char inifilename[MAX_PATH]="wcplugin.ini";  // Unused in this plugin, may be used to save data

/* Error codes returned to calling application */
#define E_END_ARCHIVE     10            /* No more files in archive */
#define E_NO_MEMORY       11            /* Not enough memory */
#define E_BAD_DATA        12            /* Data is bad */
#define E_BAD_ARCHIVE     13            /* CRC error in archive data */
#define E_UNKNOWN_FORMAT  14            /* Archive format unknown */
#define E_EOPEN           15            /* Cannot open existing file */
#define E_ECREATE         16            /* Cannot create file */
#define E_ECLOSE          17            /* Error closing file */
#define E_EREAD           18            /* Error reading from file */
#define E_EWRITE          19            /* Error writing to file */
#define E_SMALL_BUF       20            /* Buffer too small */
#define E_EABORTED        21            /* Function aborted by user */
#define E_NO_FILES        22            /* No files found */
#define E_TOO_MANY_FILES  23            /* Too many files to pack */
#define E_NOT_SUPPORTED   24            /* Function not supported */

/* flags for unpacking */
#define PK_OM_LIST         0
#define PK_OM_EXTRACT      1

/* flags for ProcessFile */
#define PK_SKIP            0            /* Skip this file */
#define PK_TEST            1            /* Test file integrity */
#define PK_EXTRACT         2            /* Extract to disk */

/* Flags passed through ChangeVolProc */
#define PK_VOL_ASK         0            /* Ask user for location of next volume */
#define PK_VOL_NOTIFY      1            /* Notify app that next volume will be unpacked */

/* Flags for packing */

/* For PackFiles */
#define PK_PACK_MOVE_FILES 1    /* Delete original after packing        */
#define PK_PACK_SAVE_PATHS 2    /* Save path names of files             */

/* Returned by GetPackCaps */
#define PK_CAPS_NEW        1    /* Can create new archives              */
#define PK_CAPS_MODIFY     2    /* Can modify exisiting archives        */
#define PK_CAPS_MULTIPLE   4    /* Archive can contain multiple files   */
#define PK_CAPS_DELETE     8    /* Can delete files                     */
#define PK_CAPS_OPTIONS   16    /* Has options dialog                   */
#define PK_CAPS_MEMPACK   32    /* Functions to pack in memory          */
#define PK_CAPS_BY_CONTENT 64   /* Can detect archive by its contents   */
#define PK_CAPS_SEARCHTEXT 128  /* Allow searching in this archive type */

#define BACKGROUND_UNPACK 1      /* Which operations are thread-safe?    */
#define BACKGROUND_PACK 2
#define BACKGROUND_MEMPACK 4

/* Options for memory packing */
#define MEM_OPTIONS_WANTHEADERS 1

/* Returned by PackToMem */
#define MEMPACK_OK         0
#define MEMPACK_DONE       1
#define MEMPACK_ERR       -1

typedef struct {
    char ArcName[260];
    char FileName[260];
    int Flags;
    int PackSize;
    int UnpSize; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tHeaderData;

typedef struct {
    char ArcName[1024];
    char FileName[1024];
    int Flags;
    DWORD PackSize;
    int PackSizeHigh;
    DWORD UnpSize; 
    int UnpSizeHigh; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
    char Unused[1024];
  } tHeaderDataEx;

typedef struct {
    WCHAR ArcNameW[1024];
    WCHAR FileNameW[1024];
    int Flags;
    DWORD PackSize;
    int PackSizeHigh;
    DWORD UnpSize; 
    int UnpSizeHigh; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
    char Unused[1024];
  } tHeaderDataExW;

typedef struct {
    char* ArcName;
    int OpenMode;
    int OpenResult;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tOpenArchiveData;

typedef struct {
    WCHAR* ArcNameW;
    int OpenMode;
    int OpenResult;
    WCHAR* CmtBufW;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tOpenArchiveDataW;

typedef struct {
	int size;
	DWORD PluginInterfaceVersionLow;
	DWORD PluginInterfaceVersionHi;
	char DefaultIniName[MAX_PATH];
} PackDefaultParamStruct;

/* Definition of callback functions called by the DLL
Ask to swap disk for multi-volume archive */
typedef int (__stdcall *tChangeVolProc)(char *ArcName,int Mode);
typedef int (__stdcall *tChangeVolProcW)(WCHAR *ArcName,int Mode);
/* Notify that data is processed - used for progress dialog */
typedef int (__stdcall *tProcessDataProc)(unsigned char *Addr,int Size);
typedef int (__stdcall *tProcessDataProcW)(WCHAR *Addr,int Size);

typedef struct {
  char ArchiveName[260];
  WCHAR ArchiveNameW[260];
  int ArchiveOpen;
  int ArchivePos;
  DWORD ArchiveSize;
  int ArchiveSizeHigh;
  DWORD UnpSize;
  int UnpSizeHigh;
  int ArchiveDate;
  tChangeVolProc ChangeVolProc;
  tChangeVolProcW ChangeVolProcW;
  tProcessDataProc ProcessDataProc;
  tProcessDataProcW ProcessDataProcW;
} ArchiveRec;

int maxarchives=10;

typedef struct {
	char data[128];
} CLzmaDec;

typedef struct {
  void *(*Alloc)(void *p, size_t size);
  void (*Free)(void *p, void *address); /* address can be 0 */
} ISzAlloc;

typedef unsigned char Byte;
typedef DWORD SizeT;

HMODULE dllinst=NULL; 

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
   	if (ul_reason_for_call==DLL_PROCESS_ATTACH) {
		dllinst=(HMODULE)hModule;
	}
	return TRUE;
}

//***************************************************************
// LZMA2 functions

#define UINT32_C(n) n ## UI32
#define UINT64_C(n) n ## UI64

#define LZMA_TELL_UNSUPPORTED_CHECK     UINT32_C(0x02)
#define LZMA_CONCATENATED               UINT32_C(0x08)

typedef enum {
	LZMA_OK                 = 0,
	LZMA_STREAM_END         = 1,
	LZMA_NO_CHECK           = 2,
	LZMA_UNSUPPORTED_CHECK  = 3,
	LZMA_GET_CHECK          = 4,
	LZMA_MEM_ERROR          = 5,
	LZMA_MEMLIMIT_ERROR     = 6,
	LZMA_FORMAT_ERROR       = 7,
	LZMA_OPTIONS_ERROR      = 8,
	LZMA_DATA_ERROR         = 9,
	LZMA_BUF_ERROR          = 10,
	LZMA_PROG_ERROR         = 11,
} lzma_ret;

typedef enum {
	LZMA_RUN = 0,
	LZMA_SYNC_FLUSH = 1,
	LZMA_FULL_FLUSH = 2,
	LZMA_FINISH = 3
} lzma_action;

typedef enum {
	LZMA_RESERVED_ENUM      = 0
} lzma_reserved_enum;

typedef enum {
	LZMA_CHECK_NONE     = 0,
	LZMA_CHECK_CRC32    = 1,
	LZMA_CHECK_CRC64    = 4,
	LZMA_CHECK_SHA256   = 10
} lzma_check;
typedef unsigned char uint8_t;
typedef __int64 uint64_t;

typedef struct {
	const uint8_t *next_in; /**< Pointer to the next input byte. */
	size_t avail_in;    /**< Number of available input bytes in next_in. */
	uint64_t total_in;  /**< Total number of bytes read by liblzma. */

	uint8_t *next_out;  /**< Pointer to the next output position. */
	size_t avail_out;   /**< Amount of free space in next_out. */
	uint64_t total_out; /**< Total number of bytes written by liblzma. */

	/**
	 * \brief       Custom memory allocation functions
	 *
	 * In most cases this is NULL which makes liblzma use
	 * the standard malloc() and free().
	 */
	void *allocator;

	/** Internal state is not visible to applications. */
	void *internal;

	/*
	 * Reserved space to allow possible future extensions without
	 * breaking the ABI. Excluding the initialization of this structure,
	 * you should not touch these, because the names of these variables
	 * may change.
	 */
	void *reserved_ptr1;
	void *reserved_ptr2;
	void *reserved_ptr3;
	void *reserved_ptr4;
	uint64_t reserved_int1;
	uint64_t reserved_int2;
	size_t reserved_int3;
	size_t reserved_int4;
	lzma_reserved_enum reserved_enum1;
	lzma_reserved_enum reserved_enum2;

} lzma_stream;

typedef __int64 uint64_t;
typedef DWORD uint32_t;
typedef unsigned char uint8_t;

typedef lzma_ret (__cdecl *tlzma_stream_decoder)(lzma_stream *strm, uint64_t memlimit, uint32_t flags);
typedef lzma_ret (__cdecl *tlzma_code)(lzma_stream *strm, lzma_action action);
typedef void (__cdecl *tlzma_end)(lzma_stream *strm);
typedef lzma_ret (__cdecl *tlzma_easy_encoder)(lzma_stream *strm, uint32_t preset, lzma_check check);

tlzma_stream_decoder lzma_stream_decoder=NULL;
tlzma_code lzma_code=NULL;
tlzma_end lzma_end=NULL;
tlzma_easy_encoder lzma_easy_encoder=NULL;

HMODULE lzmalib2=NULL;
BOOL loadtried2=false;

BOOL LoadLzma2Lib()
{
	if (!loadtried2) {
		loadtried2=true;
		char libname[260];
		// LOOK IN PLUGIN DIR
		GetModuleFileName(dllinst,libname,sizeof(libname)-8);
		char* p=strrchr(libname,'\\');
		if (p)
			p++;
		else
			p=libname;
#if defined _WIN64
		strcpy(p,"liblzma64.dll");
#else
		strcpy(p,"liblzma.dll");
#endif
		int olderrormode=SetErrorMode(0x8001);
		lzmalib2=LoadLibrary(libname);
		SetErrorMode(olderrormode);
		if (!lzmalib2)
			return false;
		lzma_stream_decoder=(tlzma_stream_decoder)GetProcAddress(lzmalib2,"lzma_stream_decoder");
		lzma_code=(tlzma_code)GetProcAddress(lzmalib2,"lzma_code");
		lzma_end=(tlzma_end)GetProcAddress(lzmalib2,"lzma_end");
		lzma_easy_encoder=(tlzma_easy_encoder)GetProcAddress(lzmalib2,"lzma_easy_encoder");

		if (lzma_stream_decoder==NULL || lzma_code==NULL || lzma_end==NULL ||
			lzma_easy_encoder==NULL) {
			FreeLibrary(lzmalib2);
			lzmalib2=NULL;
		}
	}
	return (lzmalib2!=NULL);
}


/***************************************************************/

char* strlcpy(char* p,char*p2,int maxlen)
{
	if ((int)strlen(p2)>=maxlen) {
		strncpy(p,p2,maxlen);
		p[maxlen]=0;
	} else
		strcpy(p,p2);
	return p;
}

WCHAR* wcslcpy(WCHAR* p,WCHAR*p2,int maxlen)
{
	if ((int)wcslen(p2)>=maxlen) {
		wcsncpy(p,p2,maxlen);
		p[maxlen]=0;
	} else
		wcscpy(p,p2);
	return p;
}

CRITICAL_SECTION slotcriticalsection;
BOOL slotcriticalsectioninitialized=false;

#pragma pack(1)

typedef struct {
	unsigned char xzsig[6];
	char flags[2];
	DWORD crc32;
} XZHEAD;

/* Keep a list of currently open archives (up to a maximum of maxarchives) */
static ArchiveRec ArchiveList[11];   /* 0 for packing, 1-10 for unpacking! */
                                      /* Important: array must be initialized to all 0! */
const unsigned char HEADER_MAGIC[6]
                    = { 0xFD, '7', 'z', 'X', 'Z', 0x00 };

// read encoded multi-byte number
size_t decode(unsigned char* buf, size_t size_max, __int64 *num) {
    if (size_max == 0)
        return 0;

    if (size_max > 9)
        size_max = 9;

    *num = buf[0] & 0x7F;
    size_t i = 0;

    while (buf[i++] & 0x80) {
        if (i >= size_max || buf[i] == 0x00)
            return 0;

        *num |= (__int64)(buf[i] & 0x7F) << (i * 7);
    }

    return i;
}

__int64 ReadFirstBlockHeader(int ArchiveHandle)
{
	char blockbuf[4*256];
	int n = _read ( ArchiveHandle, blockbuf, 2 );
	if (n!=2)
		return -2;
	int headsize=(blockbuf[0]+ 1) * 4;
	int filters=(blockbuf[1] & 3) + 1;

	int hascompressed=(blockbuf[1] & 0x40)!=0;
	int hasuncompressed=(blockbuf[1] & 0x80)!=0;
	if ((blockbuf[1] & 0x3C)!=0)
		return -2;   // indicates an error!
	n = _read ( ArchiveHandle, blockbuf+2, headsize-2);
	DWORD check = 0;
	check= crc32(check, (const unsigned char*)blockbuf, headsize-4);
	DWORD storedcrc=*((DWORD*)(blockbuf+headsize-4));
	if (check!=storedcrc)
		return -2;
	__int64 compressed=-1,uncompressed=-1;
	char* decodestart=blockbuf+2;
	if (hascompressed)
		decodestart+=decode((unsigned char*)decodestart, 9, &compressed);
	if (hasuncompressed)
		decodestart+=decode((unsigned char*)decodestart, 9, &uncompressed);
	return uncompressed;
}

__int64 ReadEndIndexHeader(int ArchiveHandle)
{
	// get data from end of file!
	// [4:CRC32] [4:Index Size] [2:Flags] [2:MAGIC]
	char blockbuf[12];
	lseek(ArchiveHandle,-12,SEEK_END);
	int n = _read ( ArchiveHandle, blockbuf, 12 );
	char* decodestart=blockbuf+4;
	DWORD check= crc32(0, (const unsigned char*)decodestart, 6);
	DWORD storedcrc=*((DWORD*)(&blockbuf));
	if (n!=12 || blockbuf[10]!='Y' || blockbuf[11]!='Z' || check!=storedcrc)
		return -2;
	DWORD indexsize=*((DWORD*)(blockbuf+4));
	indexsize=(indexsize+1)*4;
	lseek(ArchiveHandle,-(int)indexsize-12,SEEK_CUR);
	char* buf=(char*)malloc(indexsize);
	n = _read ( ArchiveHandle, buf, indexsize );
	if (n!=(int)indexsize|| buf[0]!=0) {
		free(buf);
		return -2;
	}
	decodestart=buf+1;
	// [numrecords] ([compressed][uncompressed])*numrecords
	__int64 numrecords;
	decodestart+=decode((unsigned char*)decodestart, 9, &numrecords);
	__int64 uncompressed=0;
	__int64 compressed=0;
	for (int i=0;i<(int)numrecords;i++) {
		__int64 size;
		decodestart+=decode((unsigned char*)decodestart, 9, &size);
		compressed+=size;
		decodestart+=decode((unsigned char*)decodestart, 9, &size);
		uncompressed+=size;
	}
	free(buf);
	return uncompressed;
}


int __stdcall OpenArchive(tOpenArchiveData* ArchiveData)
{
    int ArchiveHandle;
    int n;
    int i;
    int ArcHandle;
    int foundslot;
    unsigned short date, time;
    struct stat ftbuf;
    struct tm* fttm;
	XZHEAD xzhead;

    if (!slotcriticalsectioninitialized) {
        slotcriticalsectioninitialized=true;
        InitializeCriticalSection (&slotcriticalsection);
    }
    EnterCriticalSection (&slotcriticalsection);
    foundslot=0;
    for (i=1;i<=maxarchives;i++)
      if (ArchiveList[i].ArchiveOpen==0) {
        foundslot=1;
        ArcHandle=i;
        ArchiveList[ArcHandle].ArchiveOpen=1;
        break;
      }
    LeaveCriticalSection (&slotcriticalsection);

    if (foundslot==0) {
      ArchiveData->OpenResult=E_NO_MEMORY;
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    if((ArchiveHandle = _open(ArchiveData->ArcName,_O_BINARY | _O_RDONLY))==-1){
      ArchiveData->OpenResult=E_EOPEN;  
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }

    __int64 size64=_lseeki64(ArchiveHandle,0,SEEK_END);
	_lseek(ArchiveHandle,0,SEEK_SET);
	ArchiveList[ArcHandle].ArchiveSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].ArchiveSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    n = _read ( ArchiveHandle, &xzhead, sizeof(xzhead) );
    if (n != sizeof(xzhead)) {
      _close(ArchiveHandle);
      ArchiveData->OpenResult=E_EREAD;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }

	BOOL is_xz;
	is_xz=true;
	for (n=0;n<6;n++)
		if (xzhead.xzsig[n]!=HEADER_MAGIC[n]) {
			is_xz=false;
			break;
		}

	if (is_xz) {
		byte flags=xzhead.flags[1];
		DWORD check = 0;
		check= crc32(check, (const unsigned char*)&xzhead.flags, 2);
		if (check!=xzhead.crc32) {
		  _close(ArchiveHandle);
		  ArchiveData->OpenResult=E_BAD_ARCHIVE;
		  ArchiveList[ArcHandle].ArchiveOpen=0;
		  return 0;
		}
		size64=ReadFirstBlockHeader(ArchiveHandle);
		if (size64==-1)
			size64=ReadEndIndexHeader(ArchiveHandle);
		if (size64==-2) {
		  _close(ArchiveHandle);
		  ArchiveData->OpenResult=E_BAD_ARCHIVE;
		  ArchiveList[ArcHandle].ArchiveOpen=0;
		  return 0;
		}
	} else {
	  _close(ArchiveHandle);
      ArchiveData->OpenResult=E_BAD_ARCHIVE;
      ArchiveList[ArcHandle].ArchiveOpen=0;
	  return 0;
	}
	ArchiveList[ArcHandle].UnpSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].UnpSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    stat(ArchiveData->ArcName,&ftbuf);
    fttm=localtime(&ftbuf.st_mtime);
    time=fttm->tm_hour<<11 | fttm->tm_min<<5 | fttm->tm_sec;
    date=(fttm->tm_year-80)<<9 | (fttm->tm_mon+1)<<5 | fttm->tm_mday;
    ArchiveList[ArcHandle].ArchiveDate=65536*date+time;

    _close(ArchiveHandle);

    strcpy(ArchiveList[ArcHandle].ArchiveName,ArchiveData->ArcName);
    ArchiveList[ArcHandle].ArchiveNameW[0]=0;
    ArchiveList[ArcHandle].ArchiveOpen=1;
    ArchiveList[ArcHandle].ArchivePos=0;
    return ArcHandle;  /* returns pseudohandle */
}

int __stdcall OpenArchiveW(tOpenArchiveDataW* ArchiveData)
{
    int ArchiveHandle;
    int n;
    int i;
    int ArcHandle;
    int foundslot;
    unsigned short date, time;
    struct _stat ftbuf;
    struct tm* fttm;
	XZHEAD xzhead;

    if (!slotcriticalsectioninitialized) {
        slotcriticalsectioninitialized=true;
        InitializeCriticalSection (&slotcriticalsection);
    }
    EnterCriticalSection (&slotcriticalsection);
    foundslot=0;
    for (i=1;i<=maxarchives;i++)
      if (ArchiveList[i].ArchiveOpen==0) {
        foundslot=1;
        ArcHandle=i;
        ArchiveList[ArcHandle].ArchiveOpen=1;
        break;
      }
    LeaveCriticalSection (&slotcriticalsection);

    if (foundslot==0) {
      ArchiveData->OpenResult=E_NO_MEMORY;
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    if((ArchiveHandle = _wopen(ArchiveData->ArcNameW,_O_BINARY | _O_RDONLY))==-1){
      ArchiveData->OpenResult=E_EOPEN;  
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }

    __int64 size64=_lseeki64(ArchiveHandle,0,SEEK_END);
	_lseek(ArchiveHandle,0,SEEK_SET);
	ArchiveList[ArcHandle].ArchiveSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].ArchiveSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    n = _read ( ArchiveHandle, &xzhead, sizeof(xzhead) );
    if (n != sizeof(xzhead)) {
      _close(ArchiveHandle);
      ArchiveData->OpenResult=E_EREAD;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    
	BOOL is_xz;
	is_xz=true;
	for (n=0;n<6;n++)
		if (xzhead.xzsig[n]!=HEADER_MAGIC[n]) {
			is_xz=false;
			break;
		}

	if (is_xz) {
		DWORD check = 0;
		check= crc32(check, (const unsigned char*)&xzhead.flags, 2);
		if (check!=xzhead.crc32) {
		  _close(ArchiveHandle);
		  ArchiveData->OpenResult=E_EREAD;
		  ArchiveList[ArcHandle].ArchiveOpen=0;
		  return 0;
		}
		size64=ReadFirstBlockHeader(ArchiveHandle);
		if (size64==-1)
			size64=ReadEndIndexHeader(ArchiveHandle);
		if (size64==-2) {
		  _close(ArchiveHandle);
		  ArchiveData->OpenResult=E_BAD_ARCHIVE;
		  ArchiveList[ArcHandle].ArchiveOpen=0;
		  return 0;
		}
	} else {
	  _close(ArchiveHandle);
	  ArchiveData->OpenResult=E_BAD_ARCHIVE;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
	  return 0;
	}
    ArchiveList[ArcHandle].UnpSize=(DWORD)(size64 & 0xFFFFFFFF);
    ArchiveList[ArcHandle].UnpSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);

    _wstat(ArchiveData->ArcNameW,&ftbuf);
    fttm=localtime(&ftbuf.st_mtime);
    time=fttm->tm_hour<<11 | fttm->tm_min<<5 | fttm->tm_sec;
    date=(fttm->tm_year-80)<<9 | (fttm->tm_mon+1)<<5 | fttm->tm_mday;
    ArchiveList[ArcHandle].ArchiveDate=65536*date+time;

    _close(ArchiveHandle);

    ArchiveList[ArcHandle].ArchiveName[0]=0;
    wcscpy(ArchiveList[ArcHandle].ArchiveNameW,ArchiveData->ArcNameW);
    ArchiveList[ArcHandle].ArchiveOpen=1;
    ArchiveList[ArcHandle].ArchivePos=0;
    return ArcHandle;  /* returns pseudohandle */
}

int __stdcall CloseArchive(int hArcData)
{
    if (hArcData>0 && hArcData<=maxarchives) {
      ArchiveList[hArcData].ArchiveOpen=0;
      return 0;
    } else  
      return E_ECLOSE;
}

int __stdcall ReadHeader(int hArcData,tHeaderData* HeaderData)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        char* p;
        char* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=strrchr(ArchiveList[hArcData].ArchiveName,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveName; else p++;
        p1=strrchr(p,'.');
        if (p1 != NULL) {
          strncpy(HeaderData->FileName,p,p1-p);
        } else {
          strcpy(HeaderData->FileName,p);  
        }
        strcpy(HeaderData->ArcName,ArchiveList[hArcData].ArchiveName);
        HeaderData->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderData->UnpSize=ArchiveList[hArcData].UnpSize;
        HeaderData->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

int __stdcall ReadHeaderEx(int hArcData,tHeaderDataEx* HeaderDataEx)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        char* p;
        char* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=strrchr(ArchiveList[hArcData].ArchiveName,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveName; else p++;
        p1=strrchr(p,'.');
        if (p1 != NULL) {
          strncpy(HeaderDataEx->FileName,p,p1-p);
        } else {
          strcpy(HeaderDataEx->FileName,p);  
        }
        strcpy(HeaderDataEx->ArcName,ArchiveList[hArcData].ArchiveName);
        HeaderDataEx->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataEx->PackSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataEx->UnpSize=ArchiveList[hArcData].UnpSize;
        HeaderDataEx->UnpSizeHigh=ArchiveList[hArcData].UnpSizeHigh;
        HeaderDataEx->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

int __stdcall ReadHeaderExW(int hArcData,tHeaderDataExW* HeaderDataExW)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        WCHAR* p;
        WCHAR* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=wcsrchr(ArchiveList[hArcData].ArchiveNameW,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveNameW; else p++;
        p1=wcsrchr(p,'.');
        if (p1 != NULL) {
          wcsncpy(HeaderDataExW->FileNameW,p,p1-p);
        } else {
          wcscpy(HeaderDataExW->FileNameW,p);  
        }
        wcscpy(HeaderDataExW->ArcNameW,ArchiveList[hArcData].ArchiveNameW);
        HeaderDataExW->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataExW->PackSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataExW->UnpSize=ArchiveList[hArcData].UnpSize;
        HeaderDataExW->UnpSizeHigh=ArchiveList[hArcData].UnpSizeHigh;
        HeaderDataExW->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

typedef struct {
     FILE* handle;
} fptr;

typedef fptr* pfptr;

static BOOL myfeof ( FILE* f )
{
   int c = fgetc ( f );
   if (c == EOF) return true;
   ungetc ( c, f );
   return false;
} 

static void *SzAlloc(void *p, size_t size) { p = p; return malloc(size); }
static void SzFree(void *p, void *address) { p = p; free(address); }
static ISzAlloc g_Alloc = { SzAlloc, SzFree };

#define IN_BUF_SIZE (1 << 16)
#define OUT_BUF_SIZE (1 << 16)

#define IN_BUF_MAX	4096
#define OUT_BUF_MAX	4096
#define UINT64_MAX (UINT64_C(18446744073709551615))

static int DecodeBlockLzma2(int Operation,int hArcData,FILE *out_file, FILE *in_file)
{
	lzma_stream strm; /* alloc and init lzma_stream struct */
	memset(&strm,0,sizeof(strm));
	const uint32_t flags = LZMA_TELL_UNSUPPORTED_CHECK | LZMA_CONCATENATED;
	const uint64_t memory_limit = UINT64_MAX; /* no memory limit */
	uint8_t in_buf [IN_BUF_MAX];
	uint8_t out_buf [OUT_BUF_MAX];
	size_t in_len;	/* length of useful data in in_buf */
	size_t out_len;	/* length of useful data in out_buf */
	bool in_finished = false;
	bool out_finished = false;
	lzma_action action;
	lzma_ret ret_xz;
	int ret;

	
	int aborted=0;
	ret = 0;

	/* initialize xz decoder */
	ret_xz = lzma_stream_decoder (&strm, memory_limit, flags);
	if (ret_xz != LZMA_OK) {
		return E_NO_MEMORY;
	}


	while ((! in_finished) && (! out_finished) && !aborted) {
		/* read incoming data */
		in_len = fread (in_buf, 1, IN_BUF_MAX, in_file);

		if (feof (in_file)) {
			in_finished = true;
		}
		if (ferror (in_file)) {
			in_finished = true;
			ret = E_EREAD;
		}

		strm.next_in = in_buf;
		strm.avail_in = in_len;

		/* if no more data from in_buf, flushes the
		   internal xz buffers and closes the decompressed data
		   with LZMA_FINISH */
		action = in_finished ? LZMA_FINISH : LZMA_RUN;
		
		/* loop until there's no pending decompressed output */
		do {
			/* out_buf is clean at this point */
			strm.next_out = out_buf;
			strm.avail_out = OUT_BUF_MAX;

			/* decompress data */
			ret_xz = lzma_code (&strm, action);

			if ((ret_xz != LZMA_OK) && (ret_xz != LZMA_STREAM_END)) {
				out_finished = true;
				switch (ret_xz) {
				case LZMA_MEM_ERROR:
				case LZMA_MEMLIMIT_ERROR:
					ret = E_NO_MEMORY; break;
				case LZMA_BUF_ERROR:
					ret = E_EREAD; break;
				default:
					ret = E_BAD_DATA;
				}
			} else {
				/* write decompressed data */
				out_len = OUT_BUF_MAX - strm.avail_out;
				if (out_file!=NULL) {
					fwrite (out_buf, 1, out_len, out_file);
					if (ferror (out_file)) {
						out_finished = true;
						ret = E_EWRITE;
					}
				}
				if (Operation==PK_EXTRACT || Operation==PK_TEST) {
					if (ArchiveList[hArcData].ProcessDataProc) {
						if (ArchiveList[hArcData].ProcessDataProc(NULL,(int)(out_len))==0) {
							aborted=1;
							ret = E_EABORTED;
						}
					}
				}				  
			}
		} while (strm.avail_out == 0);

	}
	lzma_end (&strm);
	return ret;
}

int __stdcall ProcessFileAW(BOOL unicode,int hArcData,int Operation,
						   char* DestPathA,char* DestNameA,
						   WCHAR* DestPathW,WCHAR* DestNameW)
{
    if (hArcData<1 || hArcData>maxarchives)
        return E_EREAD;

    if (Operation==PK_SKIP)
        return 0;
        
    if (ArchiveList[hArcData].ArchiveOpen && ArchiveList[hArcData].ArchiveNameW[0] &&
        ((Operation==PK_EXTRACT && (DestNameA!=NULL || DestNameW !=NULL)) || (Operation==PK_TEST))) {

		if (!LoadLzma2Lib())
			return E_EREAD;

        FILE *fp_w=NULL;
		FILE *inStream;
        int aborted=0;
		int res;
		BOOL smallMode = false;
		int verbosity = 0;
		int streamNo = 0;

		if (unicode) {
			if (ArchiveList[hArcData].ProcessDataProcW) {
				if (ArchiveList[hArcData].ProcessDataProcW(DestNameW ? DestNameW:ArchiveList[hArcData].ArchiveNameW,0)==0)
					return E_EABORTED;
			}
		} else {
			if (ArchiveList[hArcData].ProcessDataProc) {
				if (ArchiveList[hArcData].ProcessDataProc((unsigned char*)(DestNameA ? DestNameA:ArchiveList[hArcData].ArchiveName),0)==0)
					return E_EABORTED;
			}
		}
        
		if (Operation==PK_EXTRACT)
			if (unicode) {
                if ((fp_w = _wfopen(DestNameW,L"wb"))==NULL)
                        return E_ECREATE;
			} else {
                if ((fp_w = fopen(DestNameA,"wb"))==NULL)
                        return E_ECREATE;
			}

		if (unicode)
			inStream = _wfopen ( ArchiveList[hArcData].ArchiveNameW, L"rb" );
		else
			inStream = fopen ( ArchiveList[hArcData].ArchiveName, "rb" );
		if ( inStream == NULL ) {
			if (Operation==PK_EXTRACT)
				fclose(fp_w);
			return E_EOPEN;
		}

		res=DecodeBlockLzma2(Operation,hArcData,fp_w,inStream);

		fclose (inStream); 

        if (Operation==PK_EXTRACT)
                fclose(fp_w);
        if (aborted)
                return E_EABORTED;
        else if (res!=0) {
			if(unicode){
				if (DestNameW)
					DeleteFileW(DestNameW);
			} else {
				if (DestNameA)
					DeleteFile(DestNameA);
			}
                return res;
        } else
                return 0;
    } else
        return E_EREAD;
}

int __stdcall ProcessFile(int hArcData,int Operation,char* DestPath,char* DestName)
{
	return ProcessFileAW(false,hArcData,Operation,DestPath,DestName,NULL,NULL);
}

int __stdcall ProcessFileW(int hArcData,int Operation,WCHAR* DestPath,WCHAR* DestName)
{
	return ProcessFileAW(true,hArcData,Operation,NULL,NULL,DestPath,DestName);
}

void __stdcall SetChangeVolProc(int hArcData,tChangeVolProc pChangeVolProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ChangeVolProc=pChangeVolProc;
}

void __stdcall SetProcessDataProc(int hArcData,tProcessDataProc pProcessDataProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ProcessDataProc=pProcessDataProc;    
}

void __stdcall SetChangeVolProcW(int hArcData,tChangeVolProcW pChangeVolProcW)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ChangeVolProcW=pChangeVolProcW;
}

void __stdcall SetProcessDataProcW(int hArcData,tProcessDataProcW pProcessDataProcW)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ProcessDataProcW=pProcessDataProcW;
}

int FileExists(char *FileName)
{
    FILE *fp_r;
    int res;
    fp_r = fopen(FileName,"rb");
    res=fp_r!=NULL;
    if (res) fclose(fp_r);
    return res;
}

int FileExistsW(WCHAR *FileName)
{
    FILE *fp_r;
    int res;
    fp_r = _wfopen(FileName,L"rb");
    res=fp_r!=NULL;
    if (res) fclose(fp_r);
    return res;
}

int RenToTemp(char *FileName,char *RenamedName)    /* returns RenamedName */
{
    char *p;
    char ch1;
    char ch2;
    strcpy(RenamedName,FileName);
    p=RenamedName;
    while (p[0]) p++;
    strncat(RenamedName,".#00",259);
    p+=2;
    for (ch1='0';ch1<='9';ch1++) {
      p[0]=ch1;
      for (ch2='0';ch2<='9';ch2++) {
        p[1]=ch2;
        if (!rename(FileName,RenamedName))
          return 1;
      }
    }
    return 0;    /* rename failed! */
}

int RenToTempW(WCHAR *FileName,WCHAR *RenamedName)    /* returns RenamedName */
{
    WCHAR *p;
    WCHAR ch1;
    WCHAR ch2;
    wcscpy(RenamedName,FileName);
    p=RenamedName;
    while (p[0]) p++;
    wcsncat(RenamedName,L".#00",259);
    p+=2;
    for (ch1='0';ch1<='9';ch1++) {
      p[0]=ch1;
      for (ch2='0';ch2<='9';ch2++) {
        p[1]=ch2;
        if (MoveFileW(FileName,RenamedName))
          return 1;
      }
    }
    return 0;    /* rename failed! */
}

/* Functions for packing files */
#define LZMA_PRESET_EXTREME       (UINT32_C(1) << 31)

int xz_compress (FILE *in_file, FILE *out_file, int level, BOOL extreme)
{
	uint32_t preset = level | (extreme ? LZMA_PRESET_EXTREME : 0);
	lzma_check check = LZMA_CHECK_CRC64;
	lzma_stream strm; /* alloc and init lzma_stream struct */
	memset(&strm,0,sizeof(strm));
	uint8_t in_buf [IN_BUF_MAX];
	uint8_t out_buf [OUT_BUF_MAX];
	size_t in_len;	/* length of useful data in in_buf */
	size_t out_len;	/* length of useful data in out_buf */
	bool in_finished = false;
	bool out_finished = false;
	lzma_action action;
	lzma_ret ret_xz;
	int ret;

	ret = 0;

	/* initialize xz encoder */
	ret_xz = lzma_easy_encoder (&strm, preset, check);
	if (ret_xz != LZMA_OK) {
		return E_NO_MEMORY;
	}

	while ((! in_finished) && (! out_finished)) {
		/* read incoming data */
		in_len = fread (in_buf, 1, IN_BUF_MAX, in_file);

		if (feof (in_file)) {
			in_finished = true;
		}
		if (ferror (in_file)) {
			in_finished = true;
			ret = E_EREAD;
		}

		strm.next_in = in_buf;
		strm.avail_in = in_len;

		/* if no more data from in_buf, flushes the
		   internal xz buffers and closes the xz data
		   with LZMA_FINISH */
		action = in_finished ? LZMA_FINISH : LZMA_RUN;

		/* loop until there's no pending compressed output */
		do {
			/* out_buf is clean at this point */
			strm.next_out = out_buf;
			strm.avail_out = OUT_BUF_MAX;

			/* compress data */
			ret_xz = lzma_code (&strm, action);

			if ((ret_xz != LZMA_OK) && (ret_xz != LZMA_STREAM_END)) {
				out_finished = true;
				ret = E_BAD_DATA;
			} else {
				/* write compressed data */
				out_len = OUT_BUF_MAX - strm.avail_out;
				fwrite (out_buf, 1, out_len, out_file);
				if (ferror (out_file)) {
					out_finished = true;
					ret = E_EWRITE;
				}
			}
		} while (strm.avail_out == 0);

		if (ArchiveList[0].ProcessDataProc) {
			if (ArchiveList[0].ProcessDataProc(NULL,(int)(in_len))==0) {
				ret = E_EABORTED;
				break;
			}
		}				  
	}

	lzma_end (&strm);
	return ret;
} 

int __stdcall PackFiles(char *PackedFile,char *SubPath,char *SrcPath,
      char *AddList,int Flags)
{
	char FullName[MAX_PATH];
	char OldName[MAX_PATH];
	FILE *fp_r,*fp_w;
	char *p;
	int retval=0;

	p=AddList;
	if (!p || !p[0]) return E_NO_FILES;
	while (p[0]) p++;
	p++;
	if (p[0]) return E_TOO_MANY_FILES;

	if (!LoadLzma2Lib())
		return E_NOT_SUPPORTED;

	// Start to pack 
	int level = GetPrivateProfileInt("xz","CompressionRate",6,inifilename);
	if (level<=0) level=1;
	if (level>=9) level=9;
	BOOL extreme = GetPrivateProfileInt("xz","Extreme",1,inifilename)!=0;

	strlcpy(FullName,SrcPath,MAX_PATH);
	strncat(FullName,AddList,MAX_PATH);

	BOOL filethere=FileExists(PackedFile);
	if (filethere) {
		// Make sure the added filename is the same as the archive name without extension!
		char *p;
		char TestName[260];
		strcpy(TestName,PackedFile);
		p=strrchr(TestName,'.');
		if (p) p[0]='\0';
		p=strrchr(TestName,'\\');
		if (p) p++;
		if (!p || stricmp(p,AddList)!=0)
				return E_TOO_MANY_FILES;
		filethere=RenToTemp(PackedFile,OldName);
	}        

	if((fp_r = fopen(FullName,"rb"))==NULL) {
		if (filethere)
			rename(OldName,PackedFile);
		return E_EOPEN;
	}        

	if ((fp_w = fopen(PackedFile,"wb"))==NULL) {
		fclose(fp_r);
		if (filethere)
			rename(OldName,PackedFile);
		return E_ECREATE;
    }
	
	retval=xz_compress(fp_r,fp_w,level,extreme);

	fclose(fp_w);
	fclose(fp_r);
	// Error occurred -> delete new archive and restore old (if exists)
	if (retval) {
		remove(PackedFile);
		if (filethere)
			rename(OldName,PackedFile);
	} else {
		if (filethere)
			unlink(OldName);
		// Move to archive - only if packing was successful!
		if (Flags & PK_PACK_MOVE_FILES)
			unlink(FullName);
	}
	return retval; 
}

int __stdcall PackFilesW(WCHAR *PackedFile,WCHAR *SubPath,WCHAR *SrcPath,
      WCHAR *AddList,int Flags)
{
	WCHAR FullName[MAX_PATH];
	WCHAR OldName[MAX_PATH];
	FILE *fp_r,*fp_w;
	WCHAR *p;
	int retval=0;

	p=AddList;
	if (!p || !p[0]) return E_NO_FILES;
	while (p[0]) p++;
	p++;
	if (p[0]) return E_TOO_MANY_FILES;

	// currently we only support packing xz files
	p=wcsrchr(PackedFile,'\\');
	if (p)
		p=wcsrchr(p,'.');
	if (p)
		if (wcsicmp(p,L".xz")!=0)
			p=NULL;
	if (!p)
		return E_NOT_SUPPORTED;

	if (!LoadLzma2Lib())
		return E_EREAD;

	// Start to pack 
	int level = GetPrivateProfileInt("xz","CompressionRate",6,inifilename);
	if (level<=0) level=1;
	if (level>=9) level=9;
	BOOL extreme = GetPrivateProfileInt("xz","Extreme",1,inifilename)!=0;

	wcslcpy(FullName,SrcPath,MAX_PATH);
	wcsncat(FullName,AddList,MAX_PATH);

	BOOL filethere=FileExistsW(PackedFile);
	if (filethere) {
		// Make sure the added filename is the same as the archive name without extension!
		WCHAR *p;
		WCHAR TestName[260];
		wcscpy(TestName,PackedFile);
		p=wcsrchr(TestName,'.');
		if (p) p[0]='\0';
		p=wcsrchr(TestName,'\\');
		if (p) p++;
		if (!p || wcsicmp(p,AddList)!=0)
				return E_TOO_MANY_FILES;
		filethere=RenToTempW(PackedFile,OldName);
	}        

	if((fp_r = _wfopen(FullName,L"rb"))==NULL) {
		if (filethere)
			MoveFileW(OldName,PackedFile);
		return E_EOPEN;
	}        

	if ((fp_w = _wfopen(PackedFile,L"wb"))==NULL) {
		fclose(fp_r);
		if (filethere)
			MoveFileW(OldName,PackedFile);
		return E_ECREATE;
    }
	
	retval=xz_compress(fp_r,fp_w,level,extreme);

	fclose(fp_w);
	fclose(fp_r);
	// Error occurred -> delete new archive and restore old (if exists)
	if (retval) {
		DeleteFileW(PackedFile);
		if (filethere)
			MoveFileW(OldName,PackedFile);
	} else {
		if (filethere)
			DeleteFileW(OldName);
		// Move to archive - only if packing was successful!
		if (Flags & PK_PACK_MOVE_FILES)
			DeleteFileW(FullName);
	}
	return retval; 
}

int __stdcall DeleteFiles(char *PackedFile,char *DeleteList)
{
  // Since this archive contains only one file anyway, just delete the whole archive!
  if (DeleteList && DeleteList[0])
    if (unlink(PackedFile))
      return E_EWRITE;
    else
      return 0;
  return E_BAD_DATA;
}

int __stdcall DeleteFilesW(WCHAR *PackedFile,WCHAR *DeleteList)
{
  // Since this archive contains only one file anyway, just delete the whole archive!
  if (DeleteList && DeleteList[0])
    if (!DeleteFileW(PackedFile))
      return E_EWRITE;
    else
      return 0;
  return E_BAD_DATA;
}

int __stdcall GetPackerCaps()
{
  return PK_CAPS_NEW | PK_CAPS_MODIFY | PK_CAPS_DELETE | PK_CAPS_OPTIONS | PK_CAPS_MEMPACK | PK_CAPS_SEARCHTEXT;
}

HANDLE hBrush;

BOOL CALLBACK DialogProc(HWND hwndDlg,UINT  uMsg,WPARAM wParam,LPARAM  lParam)
{
   int i;
   char num[4];
   char ch;
   BOOL extreme;
   num[1]='\0';
   switch (uMsg) {
     case WM_INITDIALOG:
       for (ch='1';ch<='9';ch++) {
           num[0]=ch;
           SendDlgItemMessage(hwndDlg,103,CB_ADDSTRING,0,(LPARAM)&num);
       }
       i=GetPrivateProfileInt("xz","CompressionRate",6,inifilename);
       SendDlgItemMessage(hwndDlg,103,CB_SETCURSEL,i-1,0);
	   extreme = GetPrivateProfileInt("xz","Extreme",1,inifilename)!=0;
	   if (extreme)
			CheckDlgButton(hwndDlg,105,BST_CHECKED);
       hBrush=CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
       return TRUE;
       break;
     case WM_CTLCOLORDLG:
     case WM_CTLCOLORSTATIC:
     case WM_CTLCOLORBTN:
       SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
       SetTextColor((HDC)wParam,GetSysColor(COLOR_BTNTEXT));
       return (BOOL)hBrush;
       break;
     case WM_COMMAND:
       switch (wParam) {
         case 104:    /* OK  */
           num[0]=(char)('1'+SendDlgItemMessage(hwndDlg,103,CB_GETCURSEL,0,0));
           WritePrivateProfileString("xz","CompressionRate",num,inifilename);
		   if (IsDlgButtonChecked(hwndDlg,105)==BST_CHECKED)
			   num[0]='1';
		   else
				num[0]='0';
           WritePrivateProfileString("xz","Extreme",num,inifilename);
           DeleteObject(hBrush);
           EndDialog(hwndDlg,1);
         case 2:    /* Cancel */
           DeleteObject(hBrush);
           EndDialog(hwndDlg,0);
       }
     break;
   }
   return FALSE;
}

void __stdcall ConfigurePacker(HWND ParentHandle, HINSTANCE hinstance)
{	
   /* You may ask for packing (NOT unpacking!) parameters here */
   DialogBox(hinstance,"CONFDLG",ParentHandle,(DLGPROC)&DialogProc);
}

typedef struct {
	lzma_stream strm;
} tmempackstruct,*pmempackstruct;

HANDLE __stdcall StartMemPack(int options,char* filename)
{
	if (!LoadLzma2Lib())
		return NULL;

	pmempackstruct pmps;
	pmps=(pmempackstruct)malloc(sizeof(tmempackstruct));
	if (!pmps) return 0;
	memset(pmps,0,sizeof(tmempackstruct));

	int level=max(1,min(9,GetPrivateProfileInt("xz","CompressionRate",6,inifilename)));
	BOOL extreme = GetPrivateProfileInt("xz","Extreme",1,inifilename)!=0;
	uint32_t preset = level | (extreme ? LZMA_PRESET_EXTREME : 0);
	lzma_check check = LZMA_CHECK_CRC64;

	int ret_xz = lzma_easy_encoder (&pmps->strm, preset, check);
	if (ret_xz != LZMA_OK) {
		free(pmps);
		return 0;
	}
	return (HANDLE)pmps;
}

int __stdcall PackToMem(HANDLE hMemPack,char* BufIn,int InLen,int* Taken,
      char* BufOut,int OutLen,int* Written,int* SeekBy)
{
	pmempackstruct pmps;

	pmps=(pmempackstruct)hMemPack;

	pmps->strm.next_in = (const unsigned char*)BufIn;
	pmps->strm.avail_in = InLen;
	lzma_action action = (InLen==0) ? LZMA_FINISH : LZMA_RUN;

	pmps->strm.next_out = (unsigned char*)BufOut;
	pmps->strm.avail_out = OutLen;

	/* compress data */
	int ret_xz = lzma_code (&pmps->strm, action);

	if (Written) {
		*Written=OutLen-pmps->strm.avail_out;
	}

	if (Taken) {
		*Taken=InLen-pmps->strm.avail_in;
	}
	
	if (ret_xz==LZMA_STREAM_END) {
		return MEMPACK_DONE;
	}
	else if (ret_xz==LZMA_OK)
		return MEMPACK_OK;
	else
		return MEMPACK_ERR;
}

int __stdcall DoneMemPack(HANDLE hMemPack)
{
	pmempackstruct pmps;
	if (hMemPack) {
		pmps=(pmempackstruct)hMemPack;
		lzma_end (&pmps->strm);
		free((pmempackstruct)hMemPack);
	}
	return 0;
}

int __stdcall GetBackgroundFlags(void)
{
	return BACKGROUND_UNPACK | BACKGROUND_PACK | BACKGROUND_MEMPACK;   // Support both pack+unpack+pack in memory (for tar.bz2)
}

void __stdcall PackSetDefaultParams(PackDefaultParamStruct* dps)
{
	strlcpy(inifilename,dps->DefaultIniName,MAX_PATH-1);
}


